/*
 **********************************************
 * BaseCamp: From here you can reach Everest! *
 **********************************************

 This is Deadron's BaseCamp engine for creating quests.
 The engine was inspired by Dark Age of Camelot.

 Each quest is a subclass of the BaseCamp/quest class.

 Players have a list of active quest objects and a list
 of completed quests, stored by quest type. Completed
 quests by type instead of storing the complete object
 to save space in savefiles.

 Here are the sequence of steps involved in implementing a quest
 using this library.

 --------------------------
 0) Copy an existing quest.
 --------------------------
 Seriously, start by copying.  This library is small and powerful,
 but a bit intricate. You'll really speed things up if you copy one
 of the demo quests and alter that. The Blubber quest is the simplest
 to get started with.

 I've attempted to provide a very complete explanation below of the
 library, but you'll probably figure it out better by looking at the
 code for one of the demo quests then using the doc to answer questions.

 ------------------------------------------
 1) Set your NPCs to trigger quest actions.
 ------------------------------------------
 For quests to work, they need to be given a chance to take action.
 That means the quest needs an opportunity to respond when interesting
 things happen, such as when the player clicks on an NPC.

 Typically, when the player interacts with any NPC, you'll want to
 give the player's quests a chance to do something. For example, if
 the player has killed Morbius and gotten the "Heart of the Ocean"
 jewel, if they click on the quest giver NPC, the quest should complete
 itself and give them their reward, or ask a question like "Do you have
 that jewel yet?". An NPC or object can call its base_QuestAction() proc
 to give the player's quests a chance to take action. If an action is
 taken, base_QuestAction() returns 1. In the demo, when the player clicks
 on any NPC, the NPC gives quests a chance to take action:

	mob/npc
		Click()
			// Are they close enough to talk to me?
			if (get_dist(src, usr) > 2)
				return

			// Do I have a quest-related action for this player?
			if (base_QuestAction(usr))
				return

 ----------------------------------------
 2) Set killing to trigger quest actions.
 ----------------------------------------
 Given that in typical RPGs killing a mob is part of completing many quests,
 the library has some default functionality for responding to killing by
 providing mobs with a base_PlayerKilledMob() proc that goes through the
 player's quests and calls the quest's PlayerKilledMob() proc. If you want your
 quests to respond to things besides killing, you can implement similar functions
 in your quests.

 If you want your quests to have a chance to respond to the player killing a mob,
 make sure to call the player's base_PlayerKilledMob() proc at the appropriate time.
 In the demo, when a monster is killed its Die() proc is called, and the monster
 calls base_PlayerKilledMob():

	mob/monster
		Die(mob/killer)
			// If a player killed me, let them know.
			if (killer && killer.client)
				killer.base_PlayerKilledMob(src)
			return ..()

 If you want to implement a typical "kill this kind of mob and get special loot" quest,
 all you need to do is specify the quest's victim_type and loot_type. When the quest's
 PlayerKilledMob() proc is called, the default functionality compares the victim_type to
 the dead mob type, and if they match, it'll create a loot_type object and put it in the
 player's inventory.  Typically you'll want the game to tell the player what they got,
 which you do by overriding the quest's PlayerKilledMob() proc like so:

	BaseCamp/quest
		PlayerKilledMob(mob/victim)
			var/atom/item = ..()
			if (item)
				player << "You receive the [item.name]."
			return item

 ------------------------------
 3) Define your quest subclass.
 ------------------------------
 Your quest needs to be a subclass of BaseCamp/quest, like so:

	BaseCamp/quest/blubber
		name = "Blubber Bladders Quest"

 Your subclass will contain all logic for your quest.  This is the
 secret to the library: By having all quest logic in one object,
 instead of spread out among various NPCs and objects in your game,
 you gain great power and your qeust is much easier to debug and
 maintain.

 There is more information below on how to implement your quest
 subclass.

 ---------------------------------------------------
 4) Define when the quest is available for a player.
 ---------------------------------------------------
 By default, if the player doesn't already have the quest then it's available
 unless it's a one-time-only quest and they've already done it.  If you want to
 customize this behavior, then subclass the isAvailableForPlayer() proc:

	BaseCamp/quest/blubber
		isAvailableForPlayer(mob/player)
			// 50% chance that it's available if they don't already have a Blubber quest
			if (!player.base_hasQuestOfType(type) && prob(50))
				return 1
			return 0

 --------------------------------------------
 5) Load the appropriate NPCs with the quest.
 --------------------------------------------
 To do this, add the quest's object type to the mob's base_quest_types_available list.
 This code adds the blubber quest to all NPCs in the demo:

	mob/npc
		New()
			base_quest_types_available = new()
			base_quest_types_available += /BaseCamp/quest/blubber
			return ..()

 This code adds the dagger quest to the smith, so he has both the blubber quest above
 and the dagger quest:

	mob/npc/smith
		New()
			var/result = ..()
			base_quest_types_available += /BaseCamp/quest/dagger
			return result

 -------------------------------------------------------------------
 6) Specify the first communication with the player about the quest.
 -------------------------------------------------------------------
 When a quest starts, the first thing it does is initiate dialog with the player.
 You need to override InitiateQuestDialog() to provide this.  For the blubber bladders
 quest, a web page is displayed with a link that lets the player continue the quest:

	BaseCamp/quest/blubber
		InitiateQuestDialog(atom/quest_giver)
			// Fireworks Sorcerer gives the quest.
			var/blubber_bladders = {"<a href="?command=offer_quest;src=\ref[src];npc=\ref[quest_giver]">Blubber bladders</a>"}
			rpg_browse(player, "I am in serious need of some [blubber_bladders].")

 ---------------------------------
 7) Offer the quest to the player.
 ---------------------------------
 Once the player has been told about the quest, it's time to see if they want
 to accept it. By default, the OfferQuest() proc pops up an alert asking if
 they want the quest, and responds appropriately. If you want to completely
 customize that, look at the OfferQuest implementation later on in this file.
 Otherwise, you could just use the default behavior, or enhance it.  The
 Blubber Bladders quest updates the browse window then calls the default behavior:

	BaseCamp/quest/blubber
		OfferQuest(atom/quest_giver)
			rpg_browse(player, "Don't ask me why, but I could really use some.")
			return ..()

 ----------------------------------
 8) Define the core quest behavior.
 ----------------------------------
 The QuestAction() proc is the key to how a quest works, and is completely up to you.
 It is called when the quest is accepted, and thereafter whenever the game decides to
 call the base_QuestAction() proc for an NPC (in the demo, this happens whenever a player
 clicks on an NPC). Return 1 if some action was taken, zero if not. The NPC checking for
 the quest action is passed as an argument; it's quite possible an NPC will check who has
 nothing to do with the quest. You should respond appropriately, either doing nothing and
 returning zero, or doing something and returning 1.

 If the action completes the quest, you need to call the quest's QuestCompleted() proc,
 which causes the quest to delete itself.

 See the demo quests for several examples of QuestAction().
*/

BaseCamp/quest
	var
		mob/player			// The player I belong to.
		quest_giver_type	// Object type of the mob or obj who gave the quest. We use type for persistence between sessions.

	proc/isAvailableForPlayer(mob/player)
		/*
		   Returns 1 if this quest is available for the player, zero if not.

		   The default functionality makes sure they don't already have an active
		   quest of this type, and if they aren't allowed to do the quest multiple
		   times, makes sure they haven't already completed a quest of this type.
		   Override this proc in your subclass to add your own criteria,
		   such as what level the player is.
		*/
		if (player.base_hasQuestOfType(type))
			return 0

		if (!multiple_times_allowed && player.base_completedQuestOfType(type))
			return 0
		return 1

	proc/InitiateQuestDialog(atom/quest_giver)
		/*
		   This proc needs to be overridden by your subclass.

		   The quest should ask the player if it wants the quest.
		   The typical approach is for the quest to put up a browser window,
		   with dialogue from the quest giver, using links and the Topic
		   proc to interact.

		   If the player accepts, QuestAccepted() should be called.
		   If the player rejects, QuestRejected() should be called.
		*/
		return

	proc/OfferQuest(atom/quest_giver)
		switch(alert("Do you accept the [name]?", name, "Yes", "No"))
			if ("Yes")
				QuestAccepted()
				QuestAction(quest_giver)
			else
				QuestRejected()
		return

	proc/QuestAction(atom/character)
		/*
		   This proc needs to be overridden by your subclass.

		   The character is giving the quest a chance to do something related to
		   the quest. This may be just some random character, so the subclass should
		   see if the character is related to the quest. If so, the subclass should
		   do things in the name of the character, such as saying "Bring me the Axe
		   of Murdock" or "Do you have what I asked for yet?"

		   If an action is taken, return 1; otherwise return 0.
		*/
		return

	proc/PlayerKilledMob(mob/victim)
		/*
		   Called by mob.base_PlayerKilledMob().
		   If the quest responds to this mob being killed, it should return a non-zero value.

		   If victim_type and loot_type have been set, the default functionality checks
		   to see if the mob is a victim_type; if so, creates a loot_type and adds it
		   to the player's contents, returning the item it created.  If you want to tell the
		   player what they got, override this proc with something like this:

			PlayerKilledMob(mob/victim)
				var/atom/item = ..()
				if (item)
					player << "You receive the [item.name]."
				return item

		   This is not done for you to allow you to customize the message as appropriate for your game.
		*/
		if (loot_type && (victim.type == victim_type))
			var/atom/item = new loot_type(player)
			return item
		return 0

mob
	proc/base_PlayerKilledMob(mob/victim)
		/*
		   Your game should call this when the player kills a mob.
		   It checks each quest; if a quest indicates that it has
		   responded to this (by returning 1), the function stops
		   checking, to avoid a player being rewarded in multiple
		   quests for a single kill.
		*/
		for (var/BaseCamp/quest/current_quest in base_quests_active)
			if (current_quest.PlayerKilledMob(victim))
				return 1
		return 0




/**************************************************************
 * The rest of the code you probably don't need to deal with. *
 **************************************************************/
mob
	var
		list/base_quests_active			// quest objects
		list/base_quest_types_completed	// quest types so we don't have to store complete object

	proc/base_AddQuest(BaseCamp/quest/new_quest)
		if (!base_quests_active)
			base_quests_active = new()
		base_quests_active += new_quest

	proc/base_hasQuestOfType(quest_type)
		for (var/BaseCamp/quest/current_quest in base_quests_active)
			if (current_quest.type == quest_type)
				return current_quest
		return 0

	proc/base_completedQuestOfType(quest_type)
		for (var/old_quest_type in base_quest_types_completed)
			if (old_quest_type == quest_type)
				return 1
		return 0

	proc/base_QuestCompleted(BaseCamp/quest/old_quest)
		if (!base_quest_types_completed)
			base_quest_types_completed = new()

		// Remove the type first, so we don't end up with multiple entries of one type.
		base_quest_types_completed -= old_quest.type
		base_quest_types_completed += old_quest.type

		base_RemoveQuest(old_quest)

	proc/base_RemoveQuest(BaseCamp/quest/old_quest)
		base_quests_active -= old_quest

atom
	var/tmp
		list/base_quest_types_available	// Quests I can give out

	proc/base_QuestAction(mob/player)
		// Do I have something to say to this player about an existing quest?
		for (var/BaseCamp/quest/current_quest in player.base_quests_active)
			if (current_quest.QuestAction(src))
				return 1

		// Do I have a quest to offer this player?
		for (var/quest_type in base_quest_types_available)
			var/BaseCamp/quest/current_quest = new quest_type(src)
			if (current_quest.isAvailableForPlayer(player))
				current_quest.Start(player, src)
				return 1
		return 0

BaseCamp/quest
	var/tmp
		name

		multiple_times_allowed = 1	// Can player do this quest multiple times?

		// Optional variables for default quest functionality.
		victim_type			// The type of mob the player is supposed to kill.
		loot_type			// Kind of loot player needs in contents to complete quest.
		loot_count = 1		// How much of the required_loot_type does the player need to have?

	var/global
		list/offered_quests		// Quest objects that are in the process of being accepted/rejected by players.

	proc/Start(mob/owner, atom/quest_giver)
		// Add to global list of offered quests, so I'm not garbage-collected before player accepts/rejects me.
		if (!offered_quests)
			offered_quests = new()
		offered_quests += src

		quest_giver_type = quest_giver.type
		player = owner

		InitiateQuestDialog(quest_giver)

	Topic(href, href_list[])
		var/command = href_list["command"]
		if (!command) CRASH("quest.Topic(): No \"command\" specified for quest link.")

		var/npc_ref = href_list["npc"]
		if (!npc_ref) CRASH("quest.Topic(): No \"npc\" reference specified for quest link.")

		switch(command)
			if ("offer_quest")
				var/atom/npc = locate(npc_ref)
				OfferQuest(npc)
				return
		return ..()

	proc/QuestAccepted()
		player.base_AddQuest(src)
		offered_quests -= src

	proc/QuestRejected()
		offered_quests -= src

		spawn()
			del(src)

	proc/isQuestGiver(atom/possible_giver)
		if (possible_giver.type == quest_giver_type)
			return 1
		return 0

	proc/QuestCompleted()
		player.base_QuestCompleted(src)

		// To let subclass finish up and return, don't delete myself until next cycle.
		spawn()
			del(src)

	proc/LootCount()
		if (!loot_type)	CRASH("/BaseCamp/quest/LootCount() called without loot_type being set.")

		var/count = 0
		for (var/atom/item in player.contents)
			if (item.type == loot_type)
				count++
		return count

	proc/DeleteLoot()
		// Check the player's inventory for loot_type and delete up to loot_count of them.
		if (!loot_type)	CRASH("/BaseCamp/quest/DeleteLoot() called without loot_type being set.")

		var/delete_count = 0
		for (var/atom/item in player.contents)
			if (item.type == loot_type)
				del(item)
				delete_count++
				if (delete_count == loot_count)
					break

///////////
// Tests //
///////////
/*** These tests are used during development, but aren't needed for library users.

BaseCamp/quest/test
	victim_type = /mob
	loot_type = /obj
	loot_count = 2

	New()
		// For test purposes, set up my own mob with 3 pieces of loot.
		player = new()
		for (var/count = 1, count <= 3, count++)
			new loot_type(player)
		return ..()

obj/test/verb/LootCount_test()
	var/BaseCamp/quest/test/test_quest = new()
	var/number = test_quest.LootCount()
	if (number != 3) die("LootCount() returned count of [number] instead of 3.")

obj/test/verb/DeleteLoot_test()
	// Should only delete loot_count number.
	var/BaseCamp/quest/test/test_quest = new()
	test_quest.DeleteLoot()
	var/number = test_quest.LootCount()
	if (number != 1) die("DeleteLoot() returned count of [number] instead of 1.")

obj/test/verb/PlayerKilledMob_test()
	var/BaseCamp/quest/test/test_quest = new()
	var/mob/victim = new()
	var/result = test_quest.PlayerKilledMob(victim)
	if (!result) die("PlayerKilledMob() did not recognize quest kill.")

	var/obj/false_victim = new()
	result = test_quest.PlayerKilledMob(false_victim)
	if (result) die("PlayerKilledMob() recognized wrong kill object.")

***/